import axios, { AxiosInstance, AxiosRequestConfig, AxiosResponse } from "axios";
import { ApiRequest, ApiRequestMessage, ApiResponse, ApiResponseMessage, Randomizer } from "..";
import { ApiClientError } from "./ApiClientError";
import { ApiClientErrorType } from "./ApiClientErrorType";

/**
 * 表示一个API访问客户端（跨浏览器端和NODE端）。
 */

// type ClientMockTypeFunc = (request: TRequest) => Promise<TResponse>
// interface ClientMockType {
//   [keys: string]: ClientMockTypeFunc
// }
export class ApiClient {
  // #region Constructors
  static clientMocks: {[key: string]: (request: any) => Promise<any>} = {};

  constructor(baseUrl: string, timeoutSeconds = 60000, headers?: { [key: string]: string }) {
    this.axiosInstance = axios.create({
      baseURL: baseUrl,
      timeout: timeoutSeconds * 1000,
      headers: {
        "Content-Type": "application/json",
        ...headers
      }
    });
    this.baseUrl = baseUrl;
  }

  // #endregion

  // #region methods
  /**
   * 发送API请求。
   */
  async sendAsync<TRequest extends ApiRequest, TResponse extends ApiResponse>(request: TRequest): Promise<TResponse> {
    // 该接口是否有mock
    const clientMock = ApiClient.clientMocks[request.url];
    console.log("====sendAsync===请求url=", ApiClient.clientMocks, `${this.baseUrl}${request.url}`, request);
    if (clientMock != null) {
      console.log("====sendAsync===是否有mock方法=", clientMock);
      return await clientMock(request);
    }
    // 构建Request
    const requestMessage: ApiRequestMessage<TRequest> = new ApiRequestMessage();
    requestMessage.params = request;
    const requestBody: string = JSON.stringify(requestMessage);
    const axiosRequest: AxiosRequestConfig<string> = {
      url: request.url,
      method: "POST",
      data: requestBody,
      transformRequest: (request) => request,
      transformResponse: (response) => response
    };

    // 发送Request
    let axiosResponse: AxiosResponse<string, string> | undefined;
    try {
      axiosResponse = await this.axiosInstance(axiosRequest);
    } catch {
      throw new ApiClientError(ApiClientErrorType.BadNetwork);
    }

    // 处理Response
    if (axiosResponse.status >= 300) {
      throw new ApiClientError(ApiClientErrorType.BadHttpStatus, axiosResponse.status);
    }
    const responseBody: string = axiosResponse.data;
    let responseMessage: ApiResponseMessage<TResponse> | undefined;
    try {
      responseMessage = JSON.parse(responseBody);
      if (!responseMessage) {
        throw new Error();
      }
    } catch {
      throw new ApiClientError(ApiClientErrorType.BadResponseData, axiosResponse.status);
    }
    if (!responseMessage.result || responseMessage.error) {
      throw new ApiClientError(ApiClientErrorType.Application, axiosResponse.status, responseMessage.error);
    }

    // 完成
    return responseMessage.result;
  }

  // #endregion

  // #region Implementations

  private axiosInstance: AxiosInstance;

  private baseUrl: String;
  // private clientMocks: {[key: string]: (request: any) => Promise<any>};
  // #endregion
}
